from collections import defaultdict
import json
import time
import traceback
import urllib.parse
import uuid
import grpc
import random
import requests
import logging

from requests.exceptions import ConnectionError, Timeout
from urllib3.exceptions import ProtocolError
from .loadtesting_agent import LoadtestingAgent
from yandex.cloud.loadtesting.agent.v1 import trail_service_pb2, trail_service_pb2_grpc
from yandex.cloud.loadtesting.agent.v1 import test_service_pb2_grpc, test_service_pb2, test_pb2
from yandex.cloud.loadtesting.agent.v1 import monitoring_service_pb2_grpc, monitoring_service_pb2


requests.packages.urllib3.disable_warnings()
logger = logging.getLogger(__name__)  # pylint: disable=C0103


def id_gen(base, start=0):
    i = start
    while True:
        yield '%s-%d' % (base, i)
        i += 1


class APIClient(object):
    REQUEST_ID_HEADER = 'X-Request-ID'

    def __init__(
            self,
            core_interrupted,
            base_url=None,
            writer_url=None,
            network_attempts=10,
            api_attempts=10,
            maintenance_attempts=40,
            network_timeout=2,
            api_timeout=5,
            maintenance_timeout=15,
            connection_timeout=5.0,
            user_agent=None,
            api_token=None):
        self.core_interrupted = core_interrupted
        self.user_agent = user_agent
        self.connection_timeout = connection_timeout
        self._base_url = base_url
        self.writer_url = writer_url

        self.retry_timeout = 10
        self.session = requests.Session()
        self.session.verify = False
        self.session.headers.update({"User-Agent": "tank"})

        if "https" in requests.utils.getproxies():
            logger.info("Connecting via proxy %s" % requests.utils.getproxies()['https'])
            self.session.proxies = requests.utils.getproxies()
        else:
            logger.info("Proxy not set")

        self.network_attempts = network_attempts
        self.network_timeout = network_timeout
        self.api_attempts = api_attempts
        self.api_timeout = api_timeout
        self.maintenance_attempts = maintenance_attempts
        self.maintenance_timeout = maintenance_timeout
        self.params = {'api_token': api_token} if api_token else {}

    @property
    def base_url(self):
        if not self._base_url:
            raise ValueError("Base url is not set")
        else:
            return self._base_url

    @base_url.setter
    def base_url(self, url):
        self._base_url = url

    class UnderMaintenance(Exception):
        message = "API is under maintenance"

    class NotAvailable(Exception):
        desc = "API is not available"

        def __init__(self, request, response):
            self.message = "%s\n%s\n%s" % (self.desc, request, response)
            super(self.__class__, self).__init__(self.message)

    class StoppedFromOnline(Exception):
        """http code 410"""
        message = "Shooting is stopped from online"

    class JobNotCreated(Exception):
        pass

    class NetworkError(Exception):
        pass

    def set_api_timeout(self, timeout):
        self.api_timeout = float(timeout)

    def network_timeouts(self):
        return (self.network_timeout for _ in range(self.network_attempts - 1))

    def api_timeouts(self):
        return (self.api_timeout for _ in range(self.api_attempts - 1))

    def maintenance_timeouts(self):
        return (
            self.maintenance_timeout for _ in range(
                self.maintenance_attempts - 1))

    @staticmethod
    def filter_headers(headers):
        boring = ['X-Content-Security-Policy', 'Content-Security-Policy',
                  'Strict-Transport-Security', 'X-WebKit-CSP', 'Set-Cookie',
                  'X-DNS-Prefetch-Control', 'X-Frame-Options', 'P3P',
                  'X-Content-Type-Options', 'X-Download-Options',
                  'Surrogate-Control']
        for h in boring:
            if h in headers:
                del (headers[h])
        return headers

    def __send_single_request(self, request, request_id, trace=False):
        request.headers[self.REQUEST_ID_HEADER] = request_id
        p = self.session.prepare_request(request)
        if trace:
            logger.debug(self.format_request_info(p, request_id))
        resp = self.session.send(p, timeout=self.connection_timeout)
        if trace:
            logger.debug(self.format_response_info(resp, request_id))
        if resp.status_code in [500, 502, 503, 504]:
            raise self.NotAvailable(
                request=self.format_request_info(p, request_id),
                response=self.format_response_info(resp, request_id))
        elif resp.status_code == 410:
            raise self.StoppedFromOnline
        elif resp.status_code == 423:
            raise self.UnderMaintenance
        else:
            resp.raise_for_status()
            return resp

    def format_request_info(self, request, request_id):
        utf8_body = request.body.decode('utf-8') if isinstance(request.body, bytes) else request.body
        request_info = {
            'id': request_id,
            'method': request.method,
            'url': request.url,
            'headers': str(self.filter_headers(request.headers)),
            'body': utf8_body.replace('\n', '\\n') if isinstance(utf8_body, str) else utf8_body
        }
        return """Request: {}""".format(json.dumps(request_info))

    def format_response_info(self, resp, request_id):
        response_info = {
            'id': request_id,
            'elapsed_time': resp.elapsed.total_seconds(),
            'reason': resp.reason,
            'http code': resp.status_code,
            'headers': str(self.filter_headers(resp.headers)),
            'content': resp.text.replace('\n', '\\n') if isinstance(resp.text, str) else resp.text
        }
        return """Response: {}""".format(json.dumps(response_info))

    def __make_api_request(
            self,
            http_method,
            path,
            data=None,
            response_callback=lambda x: x,
            writer=False,
            interrupted_event=None,
            trace=False,
            json=None,
            maintenance_timeouts=None,
            maintenance_msg=None):
        url = urllib.parse.urljoin(self.base_url, path)
        ids = id_gen(str(uuid.uuid4()))
        if json:
            request = requests.Request(
                http_method, url, json=json, headers={'User-Agent': self.user_agent}, params=self.params)
        else:
            request = requests.Request(
                http_method, url, data=data, headers={'User-Agent': self.user_agent}, params=self.params)
        network_timeouts = self.network_timeouts()
        maintenance_timeouts = maintenance_timeouts or self.maintenance_timeouts()
        maintenance_msg = maintenance_msg or "%s is under maintenance" % (self._base_url)
        while interrupted_event is None or not interrupted_event.is_set():
            try:
                response = self.__send_single_request(request, next(ids), trace=trace)
                return response_callback(response)
            except (Timeout, ConnectionError, ProtocolError):
                logger.warn(traceback.format_exc())
                if not self.core_interrupted.is_set():
                    try:
                        timeout = next(network_timeouts)
                        logger.warn(
                            "Network error, will retry in %ss..." %
                            timeout)
                        time.sleep(timeout)
                        continue
                    except StopIteration:
                        raise self.NetworkError()
                else:
                    break
            except self.UnderMaintenance as e:
                if not self.core_interrupted.is_set():
                    try:
                        timeout = next(maintenance_timeouts)
                        logger.warn(maintenance_msg)
                        logger.warn("Retrying in %ss..." % timeout)
                        time.sleep(timeout)
                        continue
                    except StopIteration:
                        raise e
                else:
                    break

    def __make_writer_request(
            self,
            params=None,
            json=None,
            http_method="POST",
            trace=False):
        '''
        Send request to writer service.
        '''
        request = requests.Request(
            http_method,
            self.writer_url,
            params=params,
            json=json,
            headers={
                'User-Agent': self.user_agent})
        ids = id_gen(str(uuid.uuid4()))
        network_timeouts = self.network_timeouts()
        maintenance_timeouts = self.maintenance_timeouts()
        while True:
            try:
                response = self.__send_single_request(request, next(ids), trace=trace)
                return response
            except (Timeout, ConnectionError, ProtocolError):
                logger.warn(traceback.format_exc())
                try:
                    timeout = next(network_timeouts)
                    logger.warn(
                        "Network error, will retry in %ss..." %
                        timeout)
                    time.sleep(timeout)
                    continue
                except StopIteration:
                    raise self.NetworkError()
            except self.UnderMaintenance as e:
                try:
                    timeout = next(maintenance_timeouts)
                    logger.warn(
                        "Writer is under maintenance, will retry in %ss..." %
                        timeout)
                    time.sleep(timeout)
                    continue
                except StopIteration:
                    raise e

    def __get(self, addr, trace=False, maintenance_timeouts=None, maintenance_msg=None):
        return self.__make_api_request(
            'GET',
            addr,
            trace=trace,
            response_callback=lambda r: json.loads(r.content.decode('utf8')),
            maintenance_timeouts=maintenance_timeouts,
            maintenance_msg=maintenance_msg
        )

    def __post_raw(self, addr, txt_data, trace=False, interrupted_event=None):
        return self.__make_api_request(
            'POST', addr, txt_data, lambda r: r.content, trace=trace, interrupted_event=interrupted_event)

    def __post(self, addr, data, interrupted_event=None, trace=False):
        return self.__make_api_request(
            'POST',
            addr,
            json=data,
            response_callback=lambda r: r.json(),
            interrupted_event=interrupted_event,
            trace=trace)

    def __put(self, addr, data, trace=False):
        return self.__make_api_request(
            'PUT',
            addr,
            json=data,
            response_callback=lambda r: r.text,
            trace=trace)

    def __patch(self, addr, data, trace=False):
        return self.__make_api_request(
            'PATCH',
            addr,
            json=data,
            response_callback=lambda r: r.text,
            trace=trace)

    def get_task_data(self, task, trace=False):
        return self.__get("api/task/" + task + "/summary.json", trace=trace)

    def new_job(
            self,
            task,
            person,
            tank,
            target_host,
            target_port,
            loadscheme=None,
            detailed_time=None,
            notify_list=None,
            trace=False):
        """
        :return: job_nr, upload_token
        :rtype: tuple
        """
        if not notify_list:
            notify_list = []
        data = {
            'task': task,
            'person': person,
            'tank': tank,
            'host': target_host,
            'port': target_port,
            'loadscheme': loadscheme,
            'detailed_time': detailed_time,
            'notify': notify_list
        }

        logger.debug("Job create request: %s", data)
        api_timeouts = self.api_timeouts()
        while True:
            try:
                response = self.__post(
                    "api/job/create.json", data, trace=trace)[0]
                # [{"upload_token": "1864a3b2547d40f19b5012eb038be6f6", "job": 904317}]
                return response['job'], response['upload_token']
            except (self.NotAvailable, self.StoppedFromOnline) as e:
                try:
                    timeout = next(api_timeouts)
                    logger.warn("API error, will retry in %ss..." % timeout)
                    time.sleep(timeout)
                    continue
                except StopIteration:
                    logger.warn('Failed to create job on lunapark')
                    raise self.JobNotCreated() from e
            except requests.HTTPError as e:
                raise self.JobNotCreated('Failed to create job on lunapark\n{}'.format(e.response.content))
            except Exception as e:
                logger.warn('Failed to create job on lunapark')
                logger.warn(repr(e), )
                raise self.JobNotCreated()

    def get_job_summary(self, jobno):
        result = self.__get('api/job/' + str(jobno) + '/summary.json')
        return result[0]

    def close_job(self, jobno, retcode, trace=False):
        params = {'exitcode': str(retcode)}

        result = self.__get('api/job/' + str(jobno) + '/close.json?'
                            + urllib.parse.urlencode(params), trace=trace)
        return result[0]['success']

    def edit_job_metainfo(
            self,
            jobno,
            job_name,
            job_dsc,
            instances,
            ammo_path,
            loop_count,
            version_tested,
            component,
            cmdline,
            is_starred,
            tank_type=0,
            trace=False):
        data = {
            'name': job_name,
            'description': job_dsc,
            'instances': str(instances),
            'ammo': ammo_path,
            'loop': loop_count,
            'version': version_tested,
            'component': component,
            'tank_type': int(tank_type),
            'command_line': cmdline,
            'starred': int(is_starred)
        }

        response = self.__post(
            'api/job/' + str(jobno) + '/edit.json',
            data,
            trace=trace)
        return response

    def set_imbalance_and_dsc(self, jobno, rps, comment, ts=0):
        data = {}
        if rps:
            data['imbalance'] = rps
        if comment:
            res = self.get_job_summary(jobno)
            data['description'] = (res['dsc'] + "\n" + comment).strip()

        response = self.__post('api/job/' + str(jobno) + '/edit.json', data)
        return response

    @staticmethod
    def second_data_to_push_item(data, stat, timestamp, overall, case):
        """
        @data: SecondAggregateDataItem
        """
        api_data = {
            'overall': overall,
            'case': case,
            'net_codes': [],
            'http_codes': [],
            'time_intervals': [],
            'trail': {
                'time': str(timestamp),
                'reqps': stat["metrics"]["reqps"],
                'resps': data["interval_real"]["len"],
                'expect': data["interval_real"]["total"] / 1000.0 / data["interval_real"]["len"],
                'disper': 0,
                'self_load':
                    0,  # TODO abs(round(100 - float(data.selfload), 2)),
                'input': data["size_in"]["total"],
                'output': data["size_out"]["total"],
                'connect_time': data["connect_time"]["total"] / 1000.0 / data["connect_time"]["len"],
                'send_time':
                    data["send_time"]["total"] / 1000.0 / data["send_time"]["len"],
                'latency':
                    data["latency"]["total"] / 1000.0 / data["latency"]["len"],
                'receive_time': data["receive_time"]["total"] / 1000.0 / data["receive_time"]["len"],
                'threads': stat["metrics"]["instances"],  # TODO
            }
        }

        for q, value in zip(data["interval_real"]["q"]["q"],
                            data["interval_real"]["q"]["value"]):
            api_data['trail']['q' + str(q)] = value / 1000.0

        for code, cnt in data["net_code"]["count"].items():
            api_data['net_codes'].append({'code': int(code),
                                          'count': int(cnt)})

        for code, cnt in data["proto_code"]["count"].items():
            api_data['http_codes'].append({'code': int(code),
                                           'count': int(cnt)})

        api_data['time_intervals'] = APIClient.convert_hist(data["interval_real"][
            "hist"])
        return api_data

    @staticmethod
    def convert_hist(hist):
        data = hist['data']
        bins = hist['bins']
        return [
            {
                "from": 0,  # deprecated
                "to": b / 1000.0,
                "count": count,
            } for b, count in zip(bins, data)
        ]

    def push_test_data(
            self,
            jobno,
            upload_token,
            data_item,
            stat_item,
            interrupted_event,
            trace=False):
        items = []
        uri = 'api/job/{0}/push_data.json?upload_token={1}'.format(
            jobno, upload_token)
        ts = data_item["ts"]
        for case_name, case_data in data_item["tagged"].items():
            if case_name == "":
                case_name = "__NOTAG__"
            push_item = self.second_data_to_push_item(case_data, stat_item, ts,
                                                      0, case_name)
            items.append(push_item)
        overall = self.second_data_to_push_item(data_item["overall"],
                                                stat_item, ts, 1, '')
        items.append(overall)

        api_timeouts = self.api_timeouts()
        while not interrupted_event.is_set():
            try:
                if self.writer_url:
                    res = self.__make_writer_request(
                        params={
                            "jobno": jobno,
                            "upload_token": upload_token,
                        },
                        json={
                            "trail": items,
                        },
                        trace=trace)
                    logger.debug("Writer response: %s", res.text)
                    return res.json()["success"]
                else:
                    res = self.__post(uri, items, interrupted_event, trace=trace)
                    logger.debug("API response: %s", res)
                    success = int(res[0]['success'])
                    return success
            except self.NotAvailable as e:
                try:
                    timeout = next(api_timeouts)
                    logger.warn("API error, will retry in %ss...", timeout)
                    time.sleep(timeout)
                    continue
                except StopIteration:
                    raise e

    def push_monitoring_data(
            self,
            jobno,
            upload_token,
            send_data,
            interrupted_event,
            trace=False):
        if send_data:
            addr = "api/monitoring/receiver/push?job_id=%s&upload_token=%s" % (
                jobno, upload_token)
            api_timeouts = self.api_timeouts()
            while not interrupted_event.is_set():
                try:
                    if self.writer_url:
                        res = self.__make_writer_request(
                            params={
                                "jobno": jobno,
                                "upload_token": upload_token,
                            },
                            json={
                                "monitoring": send_data,
                            },
                            trace=trace)
                        logger.debug("Writer response: %s", res.text)
                        return res.json()["success"]
                    else:
                        res = self.__post_raw(
                            addr, json.dumps(send_data), trace=trace, interrupted_event=interrupted_event)
                        logger.debug("API response: %s", res)
                        success = res == 'ok'
                        return success
                except self.NotAvailable as e:
                    try:
                        timeout = next(api_timeouts)
                        logger.warn("API error, will retry in %ss...", timeout)
                        time.sleep(timeout)
                        continue
                    except StopIteration:
                        raise e

    def push_events_data(self, jobno, operator, send_data):
        if send_data:
            # logger.info('send data: %s', send_data)
            for key in send_data:
                addr = "/api/job/{jobno}/event.json".format(
                    jobno=jobno,
                )
                body = dict(
                    operator=operator,
                    text=key[1],
                    timestamp=key[0]
                )
                api_timeouts = self.api_timeouts()
                while True:
                    try:
                        # logger.debug('Sending event: %s', body)
                        res = self.__post_raw(addr, body)
                        logger.debug("API response for events push: %s", res)
                        success = res == 'ok'
                        return success
                    except self.NotAvailable as e:
                        try:
                            timeout = next(api_timeouts)
                            logger.warn("API error, will retry in %ss...", timeout)
                            time.sleep(timeout)
                            continue
                        except StopIteration:
                            raise e

    def send_status(self, jobno, upload_token, status, trace=False):
        addr = "api/v2/jobs/%s/?upload_token=%s" % (jobno, upload_token)
        status_line = status.get("core", {}).get("stage", "unknown")
        if "stepper" in status:
            status_line += " %s" % status["stepper"].get("progress")
        api_timeouts = self.api_timeouts()
        while True:
            try:
                self.__patch(addr, {"status": status_line}, trace=trace)
                return
            except self.NotAvailable as e:
                try:
                    timeout = next(api_timeouts)
                    logger.warn("API error, will retry in %ss...", timeout)
                    time.sleep(timeout)
                    continue
                except StopIteration:
                    raise e

    def is_target_locked(self, target, trace=False):
        addr = "api/server/lock.json?action=check&address=%s" % target
        res = self.__get(addr, trace=trace)
        return res[0]

    def lock_target(self, target, duration, trace=False, maintenance_timeouts=None, maintenance_msg=None):
        addr = "api/server/lock.json?action=lock&" + \
               "address=%s&duration=%s&jobno=None" % \
               (target, int(duration))
        res = self.__get(addr, trace=trace, maintenance_timeouts=maintenance_timeouts, maintenance_msg=maintenance_msg)
        return res[0]

    def unlock_target(self, target):
        addr = self.get_manual_unlock_link(target)
        res = self.__get(addr)
        return res[0]

    def get_virtual_host_info(self, hostname):
        addr = "api/server/virtual_host.json?hostname=%s" % hostname
        res = self.__get(addr)
        try:
            return res[0]
        except KeyError:
            raise Exception(res['error'])

    @staticmethod
    def get_manual_unlock_link(target):
        return "api/server/lock.json?action=unlock&address=%s" % target

    def send_config(self, jobno, lp_requisites, config_content, trace=False):
        endpoint, field_name = lp_requisites
        logger.debug("Sending {} config".format(field_name))
        addr = "/api/job/%s/%s" % (jobno, endpoint)
        self.__post_raw(addr, {field_name: config_content}, trace=trace)

    def link_mobile_job(self, lp_key, mobile_key):
        addr = "/api/job/{jobno}/edit.json".format(jobno=lp_key)
        data = {
            'mobile_key': mobile_key
        }
        response = self.__post(addr, data)
        return response


class LPRequisites:
    CONFIGINFO = ('configinfo.txt', 'configinfo')
    MONITORING = ('jobmonitoringconfig.txt', 'monitoringconfig')
    CONFIGINITIAL = ('configinitial.txt', 'configinitial')


class OverloadClient(APIClient):
    """ mocks below for nonexistent backend methods """
    def send_status(self, jobno, upload_token, status, trace=False):
        return

    def lock_target(self, target, duration, trace=False, **kwargs):
        return

    def unlock_target(self, *args, **kwargs):
        return

    def link_mobile_job(self, lp_key, mobile_key):
        return

    def push_events_data(self, number, token, data):
        return

    def set_imbalance_and_dsc(self, jobno, rps, comment, ts=0):
        return


class CloudGRPCClient(APIClient):

    class NotAvailable(Exception):
        pass

    class AgentIdNotFound(Exception):
        pass

    def __init__(
            self,
            core_interrupted,
            loadtesting_agent: LoadtestingAgent,
            api_attempts=10,
            connection_timeout=100.0):
        super().__init__(core_interrupted)
        self.core_interrupted = core_interrupted
        self._base_url = loadtesting_agent.backend_url
        self.api_attempts = api_attempts
        self.connection_timeout = connection_timeout
        self.max_api_timeout = 120
        grpc_channel = loadtesting_agent.cloud_channel
        self.token_provider = loadtesting_agent.token_provider

        self.compute_instance_id = loadtesting_agent.compute_instance_id
        self.agent_instance_id = loadtesting_agent.agent_id
        self.trail_stub = trail_service_pb2_grpc.TrailServiceStub(grpc_channel)
        self.monitoring_stub = monitoring_service_pb2_grpc.MonitoringServiceStub(grpc_channel)
        self.test_stub = test_service_pb2_grpc.TestServiceStub(grpc_channel)

    def _request_metadata(self):
        return self.token_provider.get_auth_metadata()

    def api_timeouts(self):
        for attempt in range(self.api_attempts - 1):
            multiplier = random.uniform(1, 1.5)
            yield min(2**attempt * multiplier, self.max_api_timeout)

    @staticmethod
    def build_codes(codes_data):
        codes = []
        for code in codes_data:
            codes.append(trail_service_pb2.Trail.Codes(code=int(code['code']), count=int(code['count'])))
        return codes

    @staticmethod
    def build_intervals(intervals_data):
        intervals = []
        for interval in intervals_data:
            intervals.append(trail_service_pb2.Trail.Intervals(to=interval['to'], count=int(interval['count'])))
        return intervals

    @staticmethod
    def convert_to_proto_message(items):
        trails = []
        for item in items:
            trail_data = item["trail"]
            trail = trail_service_pb2.Trail(
                overall=int(item["overall"]),
                case_id=item["case"],
                time=trail_data["time"],
                reqps=int(trail_data["reqps"]),
                resps=int(trail_data["resps"]),
                expect=trail_data["expect"],
                input=int(trail_data["input"]),
                output=int(trail_data["output"]),
                connect_time=trail_data["connect_time"],
                send_time=trail_data["send_time"],
                latency=trail_data["latency"],
                receive_time=trail_data["receive_time"],
                threads=int(trail_data["threads"]),
                q50=trail_data.get('q50'),
                q75=trail_data.get('q75'),
                q80=trail_data.get('q80'),
                q85=trail_data.get('q85'),
                q90=trail_data.get('q90'),
                q95=trail_data.get('q95'),
                q98=trail_data.get('q98'),
                q99=trail_data.get('q99'),
                q100=trail_data.get('q100'),
                http_codes=CloudGRPCClient.build_codes(item['http_codes']),
                net_codes=CloudGRPCClient.build_codes(item['net_codes']),
                time_intervals=CloudGRPCClient.build_intervals(item['time_intervals']),
            )
            trails.append(trail)
        return trails

    def _send_data(self, func, interrupted_event, cloud_job_id, message):
        api_timeouts = self.api_timeouts()
        while not interrupted_event.is_set():
            try:
                code = func(cloud_job_id, message)
                if code == 0:
                    break
            except self.NotAvailable:
                if not self.core_interrupted.is_set():
                    try:
                        timeout = next(api_timeouts)
                    except StopIteration:
                        raise self.NotAvailable
                    logger.warning("GRPC error, will retry in %ss...", timeout)
                    time.sleep(timeout)
                    continue
                else:
                    break

    def send_trails(self, cloud_job_id, trails):
        try:
            request = trail_service_pb2.CreateTrailRequest(
                agent_instance_id=self.agent_instance_id,
                job_id=str(cloud_job_id),
                data=trails
            )
            result = self.trail_stub.Create(
                request,
                timeout=self.connection_timeout,
                metadata=self._request_metadata(),
            )
            logger.debug(f'Send trails: {trails}')
            return result.code
        except grpc.RpcError as err:
            if err.code() in (grpc.StatusCode.UNAVAILABLE, grpc.StatusCode.DEADLINE_EXCEEDED):
                raise self.NotAvailable('Connection is closed. Try to set it again.')
            raise err

    def _send_monitoring(self, cloud_job_id, chunks):
        try:
            request = monitoring_service_pb2.AddMetricRequest(
                agent_instance_id=self.agent_instance_id,
                job_id=str(cloud_job_id),
                chunks=chunks
            )
            result = self.monitoring_stub.AddMetric(
                request,
                timeout=self.connection_timeout,
                metadata=self._request_metadata(),
            )
            return result.code
        except grpc.RpcError as err:
            if err.code() in (grpc.StatusCode.UNAVAILABLE, grpc.StatusCode.DEADLINE_EXCEEDED):
                raise self.NotAvailable('Connection is closed. Try to set it again.')
            raise err

    def push_test_data(self, cloud_job_id, data_item, stat_item, interrupted_event):
        message = self.prepare_test_data(data_item, stat_item)
        self._send_data(self.send_trails, interrupted_event, cloud_job_id, message)

    @staticmethod
    def prepare_test_data(data_item, stat_item):
        items = []
        ts = data_item["ts"]
        for case_name, case_data in data_item["tagged"].items():
            if case_name == "":
                case_name = "__NOTAG__"
            push_item = CloudGRPCClient.second_data_to_push_item(case_data, stat_item, ts, 0, case_name)
            items.append(push_item)
        overall = CloudGRPCClient.second_data_to_push_item(data_item["overall"], stat_item, ts, 1, '')
        items.append(overall)
        return CloudGRPCClient.convert_to_proto_message(items)

    @staticmethod
    def _json_metric_to_proto_message(json_metrics):
        def is_float(value):
            try:
                value = float(value)
                return True
            except ValueError:
                return False

        result = []
        for metric in json_metrics:
            # handle metric as 'custom:cpu-cpu0_idle_time', using 'cpu-cpu0' as metric type, 'idle_time' as metric_name
            parts = metric.split(':')[-1].split('_')
            result.append(monitoring_service_pb2.Metric(
                metric_type=parts[0],
                metric_name='_'.join(parts[1:]),
                metric_value=float(json_metrics[metric]) if is_float(json_metrics[metric]) else 0.0,
            ))
        return result

    @staticmethod
    def _json_monitoring_data_item_to_proto_metric_chunks(monitoring_data_item):
        metrics = defaultdict(list)
        for di in monitoring_data_item:
            ts = int(di['timestamp'])
            for host_name, data in di["data"].items():
                metrics[(host_name, ts, data['comment'])].extend(
                    CloudGRPCClient._json_metric_to_proto_message(data['metrics'])
                )

        return [monitoring_service_pb2.MetricChunk(
            instance_host=key[0],
            timestamp=key[1],
            comment=key[2],
            data=chunk,
        ) for key, chunk in metrics.items()]

    def push_monitoring_data(self, cloud_job_id, monitoring_data_item, interrupted_event, trace=False):
        chunks = CloudGRPCClient._json_monitoring_data_item_to_proto_metric_chunks(monitoring_data_item)
        if trace:
            logger.debug(f'CloudGRPCClient: Send monitoring data for job ({cloud_job_id}): {chunks}')

        self._send_data(self._send_monitoring, interrupted_event, cloud_job_id, chunks)

    def create_test(self, target_address, target_port, name, description, load_schedule, config):
        schedule = test_pb2.Schedule(load_profile=str(load_schedule), load_type=1)
        request = test_service_pb2.CreateTestRequest(
            agent_instance_id=self.agent_instance_id,
            target_address=target_address,
            target_port=int(target_port),
            name=name,
            description=description,
            config=config)
        request.load_schedule.CopyFrom(schedule)
        return self.test_stub.Create(
            request,
            timeout=self.connection_timeout,
            metadata=self._request_metadata(),)

    def get_test(self, cloud_job_id):
        return self.test_stub.Get(
            test_service_pb2.GetTestRequest(test_id=cloud_job_id),
            timeout=self.connection_timeout,
            metadata=self._request_metadata()
        )

    def unlock_target(self, *args):
        return

    def set_imbalance_and_dsc(self, cloud_job_id, rps, comment, timestamp):

        request = test_service_pb2.UpdateTestRequest(
            test_id=str(cloud_job_id),
            imbalance_point=rps,
            imbalance_ts=timestamp,
            imbalance_comment=comment,
        )

        try:
            _ = self.test_stub.Update(
                request,
                timeout=self.connection_timeout,
                metadata=self._request_metadata()
            )
            logger.debug('Set imbalance %s at %s. Comment: %s', rps, timestamp, comment)
        except grpc.RpcError as err:
            if err.code() in (grpc.StatusCode.UNAVAILABLE, grpc.StatusCode.DEADLINE_EXCEEDED):
                raise self.NotAvailable('Connection is closed. Try to set it again.')
            raise err



"""
Split incoming DataFrames into chunks, cache them, union chunks with same key
and pass to the underlying aggregator.
"""

import pandas as pd


class TimeChopper(object):
    """
    TimeChopper splits incoming dataframes by index. Chunks are cached and
    chunks for same key from different DFs are joined. Then chunks are passed
    further as (<timestamp>, <dataframe>) tuples.
    """

    def __init__(self, sources):
        # self.cache_size = cache_size
        self.sources = sources
        self.cache = {}
        self.recent_ts = [0] * len(self.sources)

    def __iter__(self):
        try:
            while True:
                for n, source in enumerate(self.sources):
                    chunk = next(source)
                    if chunk is not None:
                        self.recent_ts[n] = chunk.index[-1]
                        grouped = chunk.groupby(level=0)
                        for ts, group_data in list(grouped):
                            if ts in self.cache:
                                self.cache[ts] = pd.concat(
                                    [self.cache[ts], group_data])
                            else:
                                self.cache[ts] = group_data
                last_ready_ts = min(self.recent_ts) - 1
                for ts in sorted(filter(lambda x: x <= last_ready_ts, self.cache)):
                    data = self.cache.pop(ts)
                    yield ts, data, len(data)
        except StopIteration:
            while self.cache:
                yield self.__get_result()

    def __get_result(self):
        ts = min(self.cache.keys())
        result = self.cache.pop(ts, None)
        cardinality = len(result) if result is not None else 0
        return ts, result, cardinality



import pytest

from yandextank.common.interfaces import TankInfo


class TestStatus(object):

    @pytest.mark.parametrize('updates, result', [
        ([(['plugin', 'key1'], 'foo'), (['plugin', 'key2'], 42)], {'plugin': {'key1': 'foo', 'key2': 42}}),
        ([(['plugin1', 'key1'], 'foo'),
          (['plugin1', 'key2'], 42),
          (['plugin2', 'key1'], 'bar')], {'plugin1': {'key1': 'foo', 'key2': 42},
                                          'plugin2': {'key1': 'bar'}})
    ])
    def test_update(self, updates, result):
        info = TankInfo(dict())
        for args in updates:
            info.update(*args)
        assert info.get_info_dict() == result



import logging
import json
import requests
import grpc
import time
from requests.adapters import HTTPAdapter
from typing import Tuple, Callable, Optional
import jwt
from urllib.parse import urlparse
from pathlib import Path

from yandex.cloud.iam.v1.iam_token_service_pb2 import CreateIamTokenRequest
from yandex.cloud.iam.v1.iam_token_service_pb2_grpc import IamTokenServiceStub

# ====== HELPER ======
COMPUTE_INSTANCE_METADATA_URL = 'http://169.254.169.254/computeMetadata/v1/instance/?recursive=true'
COMPUTE_INSTANCE_SA_TOKEN_URL = 'http://169.254.169.254/computeMetadata/v1/instance/service-accounts/default/token'
COMPUTE_INSTANCE_YANDEX_META_URL = 'http://169.254.169.254/computeMetadata/v1/yandex/?recursive=true'
IAM_TOKEN_SERVICE_URL = "iam.api.cloud.yandex.net:443"
TOKEN_TIMEOUT_SECONDS = 5 * 60
INFINITE_TIMEOUT_SECONDS = 5 * 365 * 24 * 60 * 60

LOGGER = logging.getLogger(__name__)


def get_instance_metadata():
    url = COMPUTE_INSTANCE_METADATA_URL
    try:
        session = requests.Session()
        session.mount(url, HTTPAdapter(max_retries=2))
        response = session.get(url, headers={"Metadata-Flavor": "Google"}).json()
        LOGGER.debug(f"Instance metadata {response}")
        return response
    except requests.exceptions.ConnectionError:
        LOGGER.warning('Compute metadata service is unavailable')
        return
    except Exception:
        LOGGER.exception("Couldn't get instance metadata of current vm")
        raise


def get_instance_yandex_metadata():
    url = COMPUTE_INSTANCE_YANDEX_META_URL
    try:
        session = requests.Session()
        session.mount(url, HTTPAdapter(max_retries=2))
        response = session.get(url, headers={"Metadata-Flavor": "Google"}).json()
        LOGGER.debug(f"Instance yandex metadata {response}")
        return response
    except requests.exceptions.ConnectionError:
        LOGGER.warning('Compute metadata service is unavailable')
        return
    except Exception:
        LOGGER.exception("Couldn't get yandex metadata of current vm")
        raise


def get_current_instance_id():
    response = get_instance_metadata()
    if response:
        return response.get('id')
    return


class AuthError(Exception):
    pass


class JWTError(AuthError):
    pass


class SAKey(object):
    def __init__(self, sa_id: str, key_id: str, key: str):
        self.sa_id = sa_id
        self.key_id = key_id
        self.key = key

    def validate(self) -> bool:
        missing_fields = [key for key, value in self.__dict__.items() if not value]
        if missing_fields:
            raise JWTError(f'All fields of SAKey must be set. Missing: {missing_fields}')

        return True


class AuthTokenProvider(object):
    def __init__(self, iam_endpoint: str = None, **kwargs) -> None:
        """There are 3 auth methods supported. It will pick one method depending on provided args:
        * iam_token: auth using this token. NOTE: after token exires, this factory is no longer able to establish connection.
                     Use only for testing purposes

        * JWT auth: auth using JWT token. This method is enabled if one following cases satisfied:
                     1. 'sa_key' of SAKey type - will use this key to generate JWT tokens.
                     2. 'sa_key' of string type, 'sa_id', 'sa_key_id' - private key in PEM format, Service Account ID and Key ID
                     3. 'sa_key_file' - path to file with service account private key (json or pem file).
                            if json file provided - it will load SAKey from that file and use for auth
                            if pem file provided - it will require 'sa_id' and 'sa_key_id' parameters

                    see https://cloud.yandex.com/en-ru/docs/iam/operations/iam-token/create-for-sa#via-jwt for more details

        * compute metadata auth: will be used if no kwargs provided; viable only for instances running in Yandex Cloud Compute VMs

        :param iam_endpoint:
            URL of custom IAM
        :param **kwargs:
            See below

        :Keyword Arguments:
            * *iam_token* (``str``) --
                use token to authorize requests
            * *sa_key* (``SAKey``) --
                Service Account key for JWT auth
            * *sa_key* (``str``) --
                Private key of Service Account key for JWT auth
            * *sa_key_file* (``str``) --
                Path to file containing Servie Account key for JWT auth
            * *sa_id* (``str``) --
                Service Account ID for JWT auth
            * *sa_key_id* (``str``) --
                Key ID for JWT auth"""

        self._iam_endpoint = iam_endpoint or IAM_TOKEN_SERVICE_URL
        self._token_requester = self.get_auth_token_requester(iam_endpoint=self._iam_endpoint, lazy_channel=self.get_iam_auth_channel, **kwargs)
        self._cached_iam_token = None
        self._expire_at = None

    def get_token(self):
        if not self._fresh():
            self._cached_iam_token, self._expire_at = self._token_requester()
            self._expire_at = min(self._expire_at, time.time() + TOKEN_TIMEOUT_SECONDS)

        return self._cached_iam_token

    def get_auth_metadata(self):
        return (("authorization", "Bearer " + self.get_token()),)

    def get_iam_auth_channel(self, options=None) -> grpc.Channel:
        return create_cloud_channel(self._iam_endpoint, channel_options=options)

    def _fresh(self):
        if self._cached_iam_token is None:
            return False
        return self._expire_at > time.time()

    @staticmethod
    def get_auth_token_requester(
        iam_endpoint: str = None,
        lazy_channel: Callable[[], grpc.Channel] = None,
        **kwargs,
    ) -> Callable[[], Tuple[str, float]]:
        iam_token = kwargs.get('iam_token', '')
        if iam_token:
            LOGGER.info('get_auth_token_requester: using iam_token')

            def iam_token_requester():
                return iam_token, time.time() + INFINITE_TIMEOUT_SECONDS
            return iam_token_requester

        sa_key = kwargs.get('sa_key', '')
        if not isinstance(sa_key, SAKey):
            sa_key = build_sa_key(**kwargs)

        if sa_key and sa_key.validate():
            LOGGER.info('get_auth_token_requester: using jwt auth')

            requester = JwtTokenRequester(
                iam_endpoint=iam_endpoint,
                sa_key=sa_key,
                channel=lazy_channel() if lazy_channel else None,
            )
            return requester.get_token

        LOGGER.info('get_auth_token_requester: compute metadata auth')
        return get_iam_token_from_metadata


def get_iam_token_from_metadata() -> Tuple[str, float]:
    url = COMPUTE_INSTANCE_SA_TOKEN_URL
    try:
        session = requests.Session()
        session.mount(url, HTTPAdapter(max_retries=2))
        token_request_time = time.time()
        raw_response = session.get(url, headers={"Metadata-Flavor": "Google"})
        response = raw_response.json()
        iam_token = response['access_token']
        expire_at = token_request_time + float(response.get('expires_in', TOKEN_TIMEOUT_SECONDS))
        LOGGER.debug("Get IAM token")
        return iam_token, expire_at
    except Exception as e:
        msg = "Couldn't get iam token for instance service account"
        LOGGER.exception(msg)
        raise AuthError(msg) from e


class JwtTokenRequester(object):
    def __init__(self, iam_endpoint: str, sa_key: SAKey, channel: grpc.Channel = None) -> None:
        self.iam_endpoint = iam_endpoint or IAM_TOKEN_SERVICE_URL
        self.sa_key = sa_key
        if not sa_key or not self.sa_key.validate():
            raise JWTError('Service account key is mandatory for JWT auth')

        channel = channel or create_cloud_channel(self.iam_endpoint)
        self.iam_stub = IamTokenServiceStub(channel)
        parsed_host, _ = _get_host_port_from_url(self.iam_endpoint)
        if not parsed_host:
            raise JWTError(f'{self.iam_endpoint} is incorrect host for JWT audience. Use "iam.api.cloud.yandex.net:443"')

        self.audience_url = f'https://{parsed_host}/iam/v1/tokens'

    def get_token(self) -> Tuple[str, float]:
        """Get an IAM token by generating and sending a JWT token to the IAM service.

        Returns a tuple of (IAM token, expiry date) or raises an exception.
        """
        try:
            response = self.iam_stub.Create(CreateIamTokenRequest(jwt=self.create_jwt()))
            return response.iam_token, response.expires_at.ToSeconds()
        except Exception as e:
            raise JWTError("Couldn't get iam token from jwt.") from e

    def create_jwt(self) -> str:
        now = int(time.time())
        payload = {
            'aud': self.audience_url,
            'iss': self.sa_key.sa_id,
            'iat': now,
            'exp': now + TOKEN_TIMEOUT_SECONDS}
        return jwt.encode(
            payload,
            self.sa_key.key,
            algorithm='PS256',
            headers={'kid': self.sa_key.key_id},
        )


def build_sa_key(**kwargs) -> Optional[SAKey]:
    sa_private_key = kwargs.get('sa_key', '')
    sa_id = kwargs.get('sa_id', '')
    sa_key_id = kwargs.get('sa_key_id', '')
    sa_key_file = kwargs.get('sa_key_file', '')
    if sa_key_file:
        sa_key = load_sa_key(sa_key_file)
        sa_key.sa_id = sa_id or sa_key.sa_id
        sa_key.key_id = sa_key_id or sa_key.key_id
        return sa_key
    elif sa_id or sa_key_id or sa_private_key:
        return SAKey(sa_id, sa_key_id, sa_private_key)

    return None


def load_sa_key(file_path: str) -> SAKey:
    try:
        return _load_sa_key_from_json(file_path)
    except json.JSONDecodeError as e:
        if e.pos > 1:
            raise JWTError('Failed to deserialize sa key file as json: incorrect json format')

    return _load_sa_key_from_pem(file_path)


def _load_sa_key_from_json(file_path: str) -> SAKey:
    with open(file_path, 'r') as f:
        key_data = json.load(f)

    return SAKey(
        key_data.get('service_account_id', ''),
        key_data.get('id', ''),
        key_data.get('private_key', '')
    )


def _load_sa_key_from_pem(file_path: str) -> SAKey:
    return SAKey(
        key_id='',
        sa_id='',
        key=Path(file_path).read_text()
    )


def _get_host_port_from_url(url: str) -> Tuple[str, str]:
    """Parse well-formed URLs and host:port pairs as well. Doesn't support ipv6 hosts."""
    result = urlparse(url)
    if not result.netloc and not url.startswith('//'):
        result = urlparse(f'//{url}')
    return result.hostname, result.port


def create_cloud_channel(backend_url, insecure_connection=False, channel_options=None) -> grpc.Channel:
    channel_options = channel_options or ()
    if insecure_connection:
        channel = grpc.insecure_channel(backend_url, channel_options + (('grpc.enable_http_proxy', 0),))
    else:
        channel = grpc.secure_channel(backend_url, grpc.ssl_channel_credentials(), channel_options)
    return channel



import mock
import paramiko

banner = '###Hellow user!####\n'


def mock_output(stream, out, banner):

    stream.read.return_value = bytes((banner + out).encode('utf-8'))

    return stream


class SSHClient(paramiko.SSHClient):

    connect = mock.Mock(return_value=None)

    def exec_command(self, command, *args, **kwargs):
        stdin = mock.MagicMock()
        stdout = mock.Mock(read=mock.Mock(return_value=b''))
        stderr = mock.Mock(read=mock.Mock(return_value=b''))

        return stdin, stdout, stderr


class SSHClientWithBanner(SSHClient):

    def exec_command(self, command, *args, **kwargs):
        stdin, stdout, stderr = super().exec_command(command, *args, **kwargs)

        if command == 'pwd':
            stdout = mock_output(stdout, '/var/tmp', banner)
        elif command == '\n':
            stdout = mock_output(stdout, '', banner)

        return stdin, stdout, stderr


class SSHClientWithoutBanner(SSHClient):

    def exec_command(self, command, *args, **kwargs):
        stdin, stdout, stderr = super().exec_command(command, *args, **kwargs)

        if command == 'pwd':
            stdout = mock_output(stdout, '/var/tmp', '')
        elif command == '\n':
            stdout = mock_output(stdout, '', '')

        return stdin, stdout, stderr



from enum import Enum
import grpc
import logging
import yaml
import os
from pathlib import Path
from dataclasses import dataclass
from typing import Optional

from yandextank.plugins.DataUploader.ycloud import get_instance_metadata, get_instance_yandex_metadata, AuthTokenProvider, build_sa_key, create_cloud_channel
from yandex.cloud.loadtesting.agent.v1 import agent_registration_service_pb2, agent_registration_service_pb2_grpc


LOGGER = logging.getLogger(__name__)  # pylint: disable=C0103

METADATA_LT_CREATED_ATTR = 'loadtesting-created'
METADATA_AGENT_VERSION_ATTR = 'agent-version'
METADATA_AGENT_NAME_ATTR = 'agent-name'
METADATA_FOLDER_ID_ATTR = 'folder-id'
YANDEX_METADATA_FOLDER_ID_ATTR = 'folderId'
ANONYMOUS_AGENT_ID = None
RUN_IN_ENVIRONMENT_ENV = 'LOADTESTING_ENVIRONMENT'


class AgentOrigin(Enum):
    UNKNOWN = 0
    COMPUTE_LT_CREATED = 1
    EXTERNAL = 2


class KnownEnvironment(Enum):
    YANDEX_COMPUTE = 'YANDEX_CLOUD_COMPUTE'


class AgentOriginError(Exception):
    pass


class LoadtestingAgent(object):
    def __init__(
        self,
        backend_url: str,
        grpc_channel: grpc.Channel,
        token_provider: AuthTokenProvider,
        agent_origin: AgentOrigin = None,
        agent_id: str = None,
        agent_id_file: str = None,
        agent_name: str = None,
        agent_version: str = None,
        folder_id: str = None,
        compute_instance_id: str = None,
        instance_lt_created: bool = False,
    ):
        self.backend_url = backend_url
        self.cloud_channel = grpc_channel
        self.token_provider = token_provider
        self.compute_instance_id = compute_instance_id
        self.instance_lt_created = bool(instance_lt_created)
        self.timeout = 30.0
        self._register_stub = agent_registration_service_pb2_grpc.AgentRegistrationServiceStub(grpc_channel)
        self.agent_id_file = agent_id_file
        self.folder_id = folder_id
        self.agent_name = agent_name
        self.agent_version = agent_version

        self.agent_origin = agent_origin or self._identify_agent_origin()
        self.agent_id = agent_id or self._identify_agent_id()

    def _identify_agent_origin(self) -> AgentOrigin:
        if self.instance_lt_created and self.compute_instance_id:
            return AgentOrigin.COMPUTE_LT_CREATED
        return AgentOrigin.EXTERNAL

    def _identify_agent_id(self) -> str:
        if self.agent_origin == AgentOrigin.COMPUTE_LT_CREATED:
            response = self._register_stub.Register(
                agent_registration_service_pb2.RegisterRequest(
                    compute_instance_id=self.compute_instance_id),
                timeout=self.timeout,
                metadata=self._request_metadata()
            )
            LOGGER.info(f'The agent has been registered with id={response.agent_instance_id}')
            return response.agent_instance_id

        if agent_id := self._load_agent_id():
            LOGGER.info(f'Load agent_id from file {agent_id}')
            return agent_id
        elif self.is_persistent_external_agent():
            args = dict(name=self.agent_name, folder_id=self.folder_id, agent_version=self.agent_version)
        elif self.is_anonymous_external_agent():
            return ANONYMOUS_AGENT_ID
        else:
            raise AgentOriginError('Unable to identify agent id. If you running external agent ensure folder id and service account key are provided')

        response = self._register_stub.ExternalAgentRegister(
            agent_registration_service_pb2.ExternalAgentRegisterRequest(
                **args
            ),
            timeout=self.timeout,
            metadata=self._request_metadata(),
        )
        metadata = agent_registration_service_pb2.ExternalAgentRegisterMetadata()
        response.metadata.Unpack(metadata)
        LOGGER.info(f'The agent has been registered with id={metadata.agent_instance_id}')
        return metadata.agent_instance_id

    def _request_metadata(self, additional_meta=None):
        meta = [(METADATA_AGENT_VERSION_ATTR, self.agent_version)] + list(self.token_provider.get_auth_metadata())
        if additional_meta:
            meta.extend(additional_meta)
        return meta

    def is_external(self) -> bool:
        return self.agent_origin == AgentOrigin.EXTERNAL

    def is_anonymous_external_agent(self) -> bool:
        return self.is_external() and not bool(self.agent_name) and self.folder_id

    def is_persistent_external_agent(self) -> bool:
        return bool(self.is_external() and self.agent_name and self.folder_id)

    def store_agent_id(self):
        if not self.agent_id:
            return
        if not self.agent_id_file:
            raise ValueError('agent_id_file parameter must be set for store_agent_id')
        with open(self.agent_id_file, 'w') as f:
            f.write(self.agent_id)

    def _load_agent_id(self) -> str:
        if self.agent_id_file:
            try:
                with open(self.agent_id_file, '+r') as f:
                    return f.read(50)
            except FileNotFoundError:
                pass

        return ''


def create_loadtesting_agent(backend_url, config=None, insecure_connection=False, channel_options=None) -> LoadtestingAgent:
    if isinstance(config, str):
        config = yaml.safe_load(Path(config).read_text())

    if not config:
        config = {}

    compute_agent_settings = try_identify_compute_metadata()
    agent_name = compute_agent_settings.agent_name or os.getenv('LOADTESTING_AGENT_NAME', config.get('agent_name'))
    folder_id = compute_agent_settings.folder_id or os.getenv('LOADTESTING_FOLDER_ID', config.get('folder_id'))
    service_account_id = os.getenv('LOADTESTING_SA_ID', config.get('service_account_id'))
    key_id = os.getenv('LOADTESTING_SA_KEY_ID', config.get('key_id'))
    private_key_file = os.getenv('LOADTESTING_SA_KEY_FILE', config.get('private_key'))
    private_key_payload = os.getenv('LOADTESTING_SA_KEY_PAYLOAD', config.get('service_account_private_key'))

    sa_key = build_sa_key(
        sa_key=private_key_payload,
        sa_key_file=private_key_file,
        sa_key_id=key_id,
        sa_id=service_account_id,
    )
    token_provider = AuthTokenProvider(
        iam_endpoint=config.get("iam_token_service_url"),
        sa_key=sa_key
    )
    cloud_channel = create_cloud_channel(backend_url, insecure_connection=insecure_connection, channel_options=channel_options)
    return LoadtestingAgent(backend_url, cloud_channel, token_provider,
                            agent_id_file=config.get('agent_id_file'),
                            agent_name=agent_name,
                            folder_id=folder_id,
                            compute_instance_id=compute_agent_settings.compute_instance_id,
                            agent_version=compute_agent_settings.agent_version,
                            instance_lt_created=compute_agent_settings.instance_lt_created)


def use_yandex_compute_metadata():
    return os.getenv(RUN_IN_ENVIRONMENT_ENV, '') == KnownEnvironment.YANDEX_COMPUTE.value


@dataclass
class _AgentComputeMetadata():
    compute_instance_id: Optional[str] = None
    agent_version: Optional[str] = None
    agent_name: Optional[str] = None
    folder_id: Optional[str] = None
    instance_lt_created: Optional[bool] = False


def try_identify_compute_metadata() -> _AgentComputeMetadata:
    if not use_yandex_compute_metadata():
        return _AgentComputeMetadata()

    metadata = get_instance_metadata()
    if not metadata:
        return _AgentComputeMetadata()

    attrs = metadata.get('attributes')
    yandex_meta = get_instance_yandex_metadata()
    agent_settings = _AgentComputeMetadata(
        compute_instance_id=metadata.get('id'),
        agent_version=attrs.get(METADATA_AGENT_VERSION_ATTR, ''),
        agent_name=attrs.get(METADATA_AGENT_NAME_ATTR, ''),
        folder_id=attrs.get(METADATA_FOLDER_ID_ATTR, yandex_meta.get(YANDEX_METADATA_FOLDER_ID_ATTR, '')),
        instance_lt_created=attrs.get(METADATA_LT_CREATED_ATTR, False),
    )
    LOGGER.info(
        'identified compute instance id "%s", agent version "%s", lt created "%s"',
        agent_settings.compute_instance_id,
        agent_settings.agent_version,
        agent_settings.instance_lt_created
    )
    return agent_settings
